Unit IntvDbg;

Interface

Uses G320,Objects,SmalFont;

Const
  GameScreenSize = 160 * 204;
  _32BIT_OPERAND = $66;
  BMnem : Array[0..15] Of String[4] =
   ('B   ','BC  ','BOV ','BPL ','BEQ ','BLT ','BLE ','BUSC',
    'NOP ','BNC ','BNOV','BMI ','BNEQ','BGE ','BGT ','BESC');
  ColorStackLoc  = $0028;
  BorderColorLoc = $002C;
  BACKTAB        = $200;
  EXEC_Loc       = $1000;
  GROM_Loc       = $3000;
  GRAM_Loc       = $3800;
  Game_Loc       = $5000;
  Boot_Loc       = $1000;

Type
  BPtr           = ^Byte;
  P1600Mem       = ^T1600Mem;
  T1600Mem       = Array[0..$FFF] Of Word;
  CP1600Mem      = Array[0..$F] Of P1600Mem;
  CP1600Regs     = Array[0..7] Of Word;
  TIOProc        = Procedure(Addr,Value: Word);
  TSpriteLoc     = Array[0..7] Of Integer;
  TPaintProc     = Procedure;
  TINTVScrenMode = (smForeBack,smColorStack);

Const
  HiRes : Boolean = False;

Var
  DebugAddr    : Word;
  Trace        : Integer;
  Debugger     : Boolean;
  ExpError     : Boolean;
  BreakExpColl : PCollection;
  VertOffset   : Integer;
  VDelay       : Integer;
  HDelay       : Integer;
  ScreenMode   : TINTVScrenMode;

Function Debug(Buf: CP1600Mem; Var Regs: CP1600Regs; Var Flags: Word;
               Var HaltLine: Integer; IOWriteProc,IOReadProc: TIOProc;
               Const SpriteX,SpriteY: TSpriteLoc;
               PaintProc: TPaintProc): Boolean;

Implementation

Uses Crt,Dos,HexWrite,SmModeX,Vectors,Equation,IntvMnu,GameStat,VGAGUI,
     MouseX;

Const
  NormTextColor = 0;
  DataTextColor = 0;
  HiTextColor   = 6;
  PCTextColor   = 2;
  HelpTextColor = 8;
  HelpHiColor   = 6;
  LineColor     = 14;
  BackColor     = 0;
  DisasmColor   = 9;
  FlagsColor    = 15;
  RegsColor     = 12;
  DataColor     = 13;
  ScrDatColor   = 2;
  TIA_Color     = 1;
  TextBGColor   = 1;
  PopupColor    = 8;
  HelpColor     = 1;
  MarkTextColor = 7;
  MarkBackColor = 0;
  ObjLocColor   = 2;

  TextStart     = 2;
  TextHeight    = 8;
  DisasmLines   = 16;
  HRDisasmLines = 32;
  DataLines     = 8;
  SmFontSize    = 1024;

  Num_Mnemonics = 78;
  Mnemonic: Array[0..Num_Mnemonics - 1] Of String[4] =
   ('HLT' ,'SDBD','EIS' ,'DIS' ,'J'   ,'JE'  ,'JD'  ,'JSR' ,'JSRE','JSRD',
    'TCI' ,'CLRC','SETC','INCR','DECR','COMR','NEGR','ADCR','GSWD','NOP',
    'SIN' ,'RSWD','SWAP','SLL' ,'RLC' ,'SLLC','SLR' ,'SAR' ,'RRC' ,'SARC',
    'JR'  ,'TSTR','MOVR','ADDR','SUBR','CMPR','ANDR','XORR','BEXT','MVO',
    'MVO@','PSHR','MVOI','MVI' ,'MVI@','PULR','MVII','ADD' ,'ADD@','ADDI',
    'SUB' ,'SUB@','SUBT','CMP' ,'CMP@','CMPI','AND' ,'AND@','ANDI','XOR',
    'XOR@','XORI','B   ','BC  ','BOV ','BPL ','BEQ ','BLT ','BLE ','BUSC',
    'NOP ','BNC ','BNOV','BMI ','BNEQ','BGE ','BGT ','BESC');

Var
  Win         : Pointer;
  BreakExp    : String;
  DataTop     : Word;

Function Assemble(St: String; PC: Word; SDBD: Boolean; Var Word1,Word2,Word3,NumWords: Word; Var Warn: Boolean): Boolean;
Var
  I       : Integer;
  P       : Integer;
  Code    : Integer;
  Code1   : Integer;
  St1     : String;
  Found   : Boolean;
  R1,R2   : Word;
  Addr    : LongInt;
  Ext     : Integer;

  Procedure Handle_SDBD;
  Begin
    If SDBD Then
    Begin
      Word2    := Addr And $FF;
      Word3    := Addr Shr 8;
      NumWords := 3;
    End
    Else
    Begin
      Word2    := Addr;
      If (Word2 And $FC00) <> 0 Then Warn := True; { More than 10 bits }
      NumWords := 2;
    End;
  End; { Handle_SDBD }

  Function Get_Reg: Boolean;
  Var Code: Integer;
  Begin
    If (St <> '') And (St[1] = 'R') Then
    Begin
      St := Trim(Copy(St,2,Length(St) - 1));
      Val(St,R1,Code);
      Get_Reg := ((Code = 0) And (R1 In [0..7]));
    End
    Else Get_Reg := False;
  End; { Get_Reg }

  Function Get_Reg_Reg: Boolean;
  Var P,Code,Code1: Integer;
  Begin
    If (St <> '') And (St[1] = 'R') Then
    Begin
      P := Pos(',',St);
      If P <> 0 Then
      Begin
        St1 := Trim(Copy(St,2,P - 2));
        St  := Trim(Copy(St,P + 1,Length(St) - P));
        If (St <> '') And (St[1] = 'R') Then
        Begin
          St := Trim(Copy(St,2,Length(St) - 1));
          Val(St1,R1,Code);
          Val(St,R2,Code1);
          Get_Reg_Reg := ((Code  = 0) And (R1 In [0..7]) And
                          (Code1 = 0) And (R2 In [0..7]));
        End
        Else Get_Reg_Reg := False;
      End
      Else Get_Reg_Reg := False;
    End
    Else Get_Reg_Reg := False;
  End; { Get_Reg_Reg }

  Function Get_Reg_Word: Boolean;
  Var P,Code,Code1: Integer;
  Begin
    If (St <> '') And (St[1] = 'R') Then
    Begin
      P := Pos(',',St);
      If P <> 0 Then
      Begin
        St1 := Trim(Copy(St,2,P - 2));
        St  := Trim(Copy(St,P + 1,Length(St) - P));
        Val(St1,R1,Code);
        Addr := ValHex(St);
        Get_Reg_Word := ((Code = 0) And (R1 In [0..7]) And
                         (Addr >= 0) And (Addr <= $FFFF));
      End
      Else Get_Reg_Word := False;
    End
    Else Get_Reg_Word := False;
  End; { Get_Reg_Word }

  Function Get_Word_Reg: Boolean;
  Var P,Code,Code1: Integer;
  Begin
    If St <> '' Then
    Begin
      P := Pos(',',St);
      If P <> 0 Then
      Begin
        St1 := Trim(Copy(St,1,P - 1));
        St  := Trim(Copy(St,P + 1,Length(St) - P));
        If (St <> '') And (St[1] = 'R') Then
        Begin
          St := Trim(Copy(St,2,Length(St) - 1));
          Addr := ValHex(St1);
          Val(St,R1,Code);
          Get_Word_Reg := ((R1 In [0..7]) And (Code = 0) And
                           (Addr >= 0) And (Addr <= $FFFF));
        End
        Else Get_Word_Reg := False;
      End
      Else Get_Word_Reg := False;
    End
    Else Get_Word_Reg := False;
  End; { Get_Word_Reg }

Begin

  { Separate the mnemonic from the parameters }

  Warn := False;
  St   := UpperCase(Trim(St));
  If St <> '' Then
  Begin
    I     := Pos(' ',St);
    If I = 0 Then I := Length(St) + 1;
    St1   := Copy(St,1,I - 1);
    St    := Trim(Copy(St,I + 1,Length(St) - I));

    { Identify the mnemonic }

    I     := 0;
    Found := False;
    While (I < Num_Mnemonics) And Not Found Do
    Begin
      If St1 = Mnemonic[I] Then Found := True Else Inc(I);
    End; { While }
    If Found Then
    Begin
      Assemble := False;
      Case I Of
        0..3: Begin              { HLT, SDBD, EIS, DIS }
                Word1    := I;
                NumWords := 1;
                Assemble := True;
              End;
       4,5,6: Begin              { J, JE, JD }
                Addr     := ValHex(St);
                If (Addr >= 0) And (Addr <= $FFFF) Then
                Begin
                  Word1    := 4;
                  Word2    := $0300 + ((Addr And $FC00) Shr 8) + (I - 4);
                  Word3    := Addr And $3FF;
                  NumWords := 3;
                  Assemble := True;
                End;
              End;
       7,8,9: Begin              { JSR, JSRE, JSRD }
                P := Pos(',',St);
                If (P <> 0) And (St[1] = 'R') Then
                Begin
                  St1 := Trim(Copy(St,2,P - 2)); { Register number }
                  St  := Trim(Copy(St,P + 1,Length(St) - P)); { Address }
                  Val(St1,R1,Code);
                  If (Code = 0) And (R1 In [4..6]) Then
                  Begin
                    Addr     := ValHex(St);
                    If (Addr >= 0) And (Addr <= $FFFF) Then
                    Begin
                      Word1    := 4;
                      Word2    := ((R1 - 4) Shl 8) + ((Addr And $FC00) Shr 8) + (I - 7);
                      Word3    := Addr And $3FF;
                      NumWords := 3;
                      Assemble := True;
                    End;
                  End;
                End;
              End;
      10..12: Begin              { TCI, CLRC, SETC }
                Word1    := I - 5;
                NumWords := 1;
                Assemble := True;
              End;
      13..17: Begin              { INCR, DECR, COMR, NEGR, ADCR }
                If Get_Reg Then
                Begin
                  Word1    := R1 + (I - 12) * 8;
                  NumWords := 1;
                  Assemble := True;
                End;
              End;
          18: Begin              { GSWD }
                If Get_Reg And (R1 In [0..3]) Then
                Begin
                  Word1    := R1 + 48;
                  NumWords := 1;
                  Assemble := True;
                End;
              End;
       19,20: Begin              { NOP,SIN }
                Word1    := 52 + (I - 19) * 2;
                NumWords := 1;
                Assemble := True;
              End;
          21: Begin              { RSWD }
                If Get_Reg Then
                Begin
                  Word1    := R1 + 56;
                  NumWords := 1;
                  Assemble := True;
                End;
              End;
      22..29: Begin              { SWAP, SLL, RLC, SLLC, SLR, SAR, RRC, SARC }
                If (St <> '') And (St[1] = 'R') Then
                Begin
                  P := Pos(',2',St);
                  If P <> 0                              { Register number }
                   Then St := Trim(Copy(St,2,P - 2))
                   Else St := Trim(Copy(St,2,Length(St) - 1));
                  Val(St,R1,Code);
                  If (Code = 0) And (R1 In [0..3]) Then
                  Begin
                    Word1 := 64 + (I - 22) * 8 + R1;
                    If P <> 0 Then Word1 := Word1 Or 4;
                    NumWords := 1;
                    Assemble := True;
                  End;
                End;
              End;
          30: Begin              { JR (MOVR RR,R7) }
                If Get_Reg Then
                Begin
                  Word1    := $0087 + (R1 Shl 3);
                  NumWords := 1;
                  Assemble := True;
                End;
              End;
          31: Begin              { TSTR (MOVR RR,RR) }
                If Get_Reg Then
                Begin
                  Word1    := $0080 + (R1 Shl 3) + R1;
                  NumWords := 1;
                  Assemble := True;
                End;
              End;
      32..37: Begin              { MOVR, ADDR, SUBR, CMPR, ANDR, XORR }
                If Get_Reg_Reg Then
                Begin
                  Word1    := $0080 + ($40 * (I - 32)) + (R1 Shl 3) + R2;
                  NumWords := 1;
                  Assemble := True;
                End;
              End;
          38: Begin              { BEXT }
                P := Pos(',',St);
                If P <> 0 Then
                Begin
                  St1  := Trim(Copy(St,1,P - 1));
                  St   := Trim(Copy(St,P + 1,Length(St) - P));
                  Addr := ValHex(St1);
                  Ext  := ValHex(St);
                  If (Ext  >= 0) And (Ext <= $F) And
                     (Addr >= 0) And (Addr <= $FFFF) Then
                  Begin
                    If Addr >= PC + 2 Then
                    Begin
                      Dec(Addr,PC + 2);
                      Word1 := $0210 + Ext;
                    End
                    Else
                    Begin
                      Addr  := PC + 1 - Addr;
                      Word1 := $0230 + Ext;
                    End;
                    Word2    := Addr;
                    If (Word2 And $FC00) <> 0 Then Warn := True; { More than 10 bits }
                    NumWords := 2;
                    Assemble := True;
                  End;
                End;
              End;
          39: Begin              { MVO }
                If Get_Reg_Word Then
                Begin
                  Word1 := $0240 + (R1 Shl 3);
                  Handle_SDBD;
                  Assemble := True;
                End;
              End;
          40: Begin              { MVO@ }
                If Get_Reg_Reg Then
                Begin
                  Word1    := $0240 + (R2 Shl 3) + R1;
                  NumWords := 1;
                  Assemble := True;
                End;
              End;
          41: Begin              { PSHR (MVO@ RR,R6) }
                If Get_Reg Then
                Begin
                  Word1    := $0270 + R1;
                  NumWords := 1;
                  Assemble := True;
                End;
              End;
          42: Begin              { MVOI (MVO@ RR,R7) }
                If Get_Reg_Word Then
                Begin
                  Word1 := $0278 + R1;
                  Handle_SDBD;
                  Assemble := True;
                End;
              End;
          43: Begin              { MVI }
                If Get_Word_Reg Then
                Begin
                  Word1 := $0280 + (R1 Shl 3);
                  Handle_SDBD;
                  Assemble := True;
                End;
              End;
          44: Begin              { MVI@ }
                If Get_Reg_Reg Then
                Begin
                  Word1    := $0280 + (R1 Shl 3) + R2;
                  NumWords := 1;
                  Assemble := True;
                End;
              End;
          45: Begin              { PULR (MVI@ R6,RR) }
                If Get_Reg Then
                Begin
                  Word1    := $02B0 + R1;
                  NumWords := 1;
                  Assemble := True;
                End;
              End;
          46: Begin              { MVII (MVI@ R7,RR) }
                If Get_Word_Reg Then
                Begin
                  Word1 := $02B8 + R1;
                  Handle_SDBD;
                  Assemble := True;
                End;
              End;
  47,50,53,56,59: Begin          { ADD, SUB, CMP, AND, XOR }
                If Get_Word_Reg Then
                Begin
                  Word1 := $02C0 + ((I - 47) Div 3) * $40 + R1;
                  Handle_SDBD;
                  Assemble := True;
                End;
              End;
  48,51,54,57,60: Begin          { ADD@, SUB@, CMP@, AND@, XOR@ }
                If Get_Reg_Reg Then
                Begin
                  Word1    := $02C0 + ((I - 48) Div 3) * $40 + (R1 Shl 3) + R2;
                  NumWords := 1;
                  Assemble := True;
                End;
              End;
  49,52,55,58,61: Begin          { ADDI, SUBT, CMPI, ANDI, XORI }
                If Get_Word_Reg Then
                Begin
                  Word1 := $02F8 + R1 + ((I - 49) Div 3) * $40;
                  Handle_SDBD;
                  Assemble := True;
                End;
              End;
      62..78: Begin              { Conditional branches }
                Addr := ValHex(St);
                If (Addr >= 0) And (Addr <= $FFFF) Then
                Begin
                  If Addr >= PC + 2 Then
                  Begin
                    Dec(Addr,PC + 2);
                    Word1 := $0200 + I - 62;
                  End
                  Else
                  Begin
                    Addr  := PC + 1 - Addr;
                    Word1 := $0220 + I - 62;
                  End;
                  Word2    := Addr;
                  If (Word2 And $FC00) <> 0 Then Warn := True; { More than 10 bits }
                  NumWords := 2;
                  Assemble := True;
                End;
              End;
      End; { Case }
    End
    Else Assemble := False;
  End
  Else Assemble := False;
End; { Assemble }

Function Disassemble(Buf: CP1600Mem; Var Addr: Word; SDBD: Boolean): String;
Var
  St      : String;
  J       : Word;
  Disp    : LongInt;
  Opcode  : Word;
  Opcode2 : Word;
  Opcode3 : Word;

  Procedure Disasm_ALU_Mem(Base: Word; Const Mnem1,Mnem2,Mnem3: String);
  Begin
    If Opcode < Base + 8 Then
    Begin
      If SDBD Then
      Begin
        St := Mnem1 + '   ' + HexWord((Opcode2 And $FF) + (Opcode3 Shl 8)) +
                       ',R' + IntToStr(Opcode And 7);
        Inc(Addr,2);
      End
      Else
      Begin
        St := Mnem1 + '   ' + HexWord(Opcode2) +
                       ',R' + IntToStr(Opcode And 7);
        Inc(Addr);
      End;
    End
    Else
    Begin
      If ((Opcode Shr 3) And 7) <> 7 Then
       St := Mnem2 + '   R' + IntToStr((Opcode Shr 3) And 7) +
                       ',R' + IntToStr(Opcode And 7)
      Else
      Begin
        If SDBD Then
        Begin
          St := Mnem3 + '   ' + HexWord((Opcode2 And $FF) + (Opcode3 Shl 8)) +
                         ',R' + IntToStr(Opcode And 7);
          Inc(Addr,2);
        End
        Else
        Begin
          St := Mnem3 + '   ' + HexWord(Opcode2) +
                         ',R' + IntToStr(Opcode And 7);
          Inc(Addr);
        End;
      End;
    End;
  End; { Disasm_ALU_Mem }

Begin
  St  := '';
  Opcode  := Buf[Addr Shr 12]^[Addr And $FFF] And $3FF;
  Opcode2 := Buf[(Addr + 1) Shr 12]^[(Addr + 1) And $FFF] And $3FF;
  Opcode3 := Buf[(Addr + 2) Shr 12]^[(Addr + 2) And $FFF] And $3FF;
  Case Opcode Of
     0: St := 'HLT';
     1: St := 'SDBD';
     2: St := 'EIS';
     3: St := 'DIS';
     4: Begin
          Case (Opcode2 And 3) Of
            0: If (Opcode2 And $300) = $300 Then St := 'J      ' Else St := 'JSR    R';
            1: If (Opcode2 And $300) = $300 Then St := 'JE     ' Else St := 'JSRE   R';
            2: If (Opcode2 And $300) = $300 Then St := 'JD     ' Else St := 'JSRD   R';
          End; { Case }
          If (Opcode2 And $300) <> $300 Then St := St + IntToStr(((Opcode2 And $300) Shr 8) + 4) + ',';
          St := St + HexWord((Opcode3 And $3FF) + ((Opcode2 And $FC) Shl 8));
          Inc(Addr,2);
        End;
     5: St := 'TCI';
     6: St := 'CLRC';
     7: St := 'SETC';
 8..15: St := 'INCR   R' + IntToStr(Opcode And 7);
16..23: St := 'DECR   R' + IntToStr(Opcode And 7);
24..31: St := 'COMR   R' + IntToStr(Opcode And 7);
32..39: St := 'NEGR   R' + IntToStr(Opcode And 7);
40..47: St := 'ADCR   R' + IntToStr(Opcode And 7);
48..51: St := 'GSWD   R' + IntToStr(Opcode And 3);
52..53: St := 'NOP';
54..55: St := 'SIN';
56..63: St := 'RSWD   R' + IntToStr(Opcode And 7);
64..71: Begin
          St := 'SWAP   R' + IntToStr(Opcode And 3);
          If (Opcode And 4) <> 0 Then St := St + ',2';
        End;
72..79: Begin
          St := 'SLL    R' + IntToStr(Opcode And 3);
          If (Opcode And 4) <> 0 Then St := St + ',2';
        End;
80..87: Begin
          St := 'RLC    R' + IntToStr(Opcode And 3);
          If (Opcode And 4) <> 0 Then St := St + ',2';
        End;
88..95: Begin
          St := 'SLLC   R' + IntToStr(Opcode And 3);
          If (Opcode And 4) <> 0 Then St := St + ',2';
        End;
96..103: Begin
          St := 'SLR    R' + IntToStr(Opcode And 3);
          If (Opcode And 4) <> 0 Then St := St + ',2';
        End;
104..111: Begin
          St := 'SAR    R' + IntToStr(Opcode And 3);
          If (Opcode And 4) <> 0 Then St := St + ',2';
        End;
112..119: Begin
          St := 'RRC    R' + IntToStr(Opcode And 3);
          If (Opcode And 4) <> 0 Then St := St + ',2';
        End;
120..127: Begin
          St := 'SARC   R' + IntToStr(Opcode And 3);
          If (Opcode And 4) <> 0 Then St := St + ',2';
        End;
128..191: Begin
          If (Opcode And 7) = 7 Then
           St := 'JR     R' + IntToStr((Opcode Shr 3) And 7)
          Else
          Begin
            If (Opcode And 7) = ((Opcode Shr 3) And 7)
             Then St := 'TSTR   R' + IntToStr((Opcode Shr 3) And 7)
             Else St := 'MOVR   R' + IntToStr((Opcode Shr 3) And 7) +
                              ',R' + IntToStr(Opcode And 7);
          End;
        End;
192..255: St := 'ADDR   R' + IntToStr((Opcode Shr 3) And 7) +
                      ',R' + IntToStr(Opcode And 7);
256..319: St := 'SUBR   R' + IntToStr((Opcode Shr 3) And 7) +
                      ',R' + IntToStr(Opcode And 7);
320..383: St := 'CMPR   R' + IntToStr((Opcode Shr 3) And 7) +
                      ',R' + IntToStr(Opcode And 7);
384..447: St := 'ANDR   R' + IntToStr((Opcode Shr 3) And 7) +
                      ',R' + IntToStr(Opcode And 7);
448..511: St := 'XORR   R' + IntToStr((Opcode Shr 3) And 7) +
                      ',R' + IntToStr(Opcode And 7);
512..575: Begin
          Disp := Opcode2;
          If (Opcode And $20) <> 0 Then Disp := -(Disp + 1);
          Inc(Disp,Addr + 2);
          If (Opcode And $10) <> 0
           Then St := 'BEXT   ' + HexWord(Disp) +
                            ',' + HexByte(Opcode And $F)
           Else St := BMnem[Opcode And $F] + '   ' + HexWord(Disp);
          Inc(Addr);
        End;
576..583: Begin
          If SDBD Then
          Begin
            St := 'MVO    R' + IntToStr(Opcode And 7) +
                         ',' + HexWord((Opcode2 And $FF) + (Opcode3 Shl 8));
            Inc(Addr,2);
          End
          Else
          Begin
            St := 'MVO    R' + IntToStr(Opcode And 7) +
                         ',' + HexWord(Opcode2);
            Inc(Addr);
          End;
        End;
584..639: Begin
          Case (Opcode Shr 3) And 7 Of
            0..5: St := 'MVO@   R' + IntToStr(Opcode And 7) +
                              ',R' + IntToStr((Opcode Shr 3) And 7);
               6: St := 'PSHR   R' + IntToStr(Opcode And 7);
               7: Begin
                    If SDBD Then
                    Begin
                      St := 'MVOI   R' + IntToStr(Opcode And 7) +
                                   ',' + HexWord((Opcode2 And $FF) + (Opcode3 Shl 8));
                      Inc(Addr,2);
                    End
                    Else
                    Begin
                      St := 'MVOI   R' + IntToStr(Opcode And 7) +
                                   ',' + HexWord(Opcode2);
                      Inc(Addr);
                    End;
                  End;
          End; { Case }
        End;
640..647: Begin
          If SDBD Then
          Begin
            St := 'MVI    ' + HexWord((Opcode2 And $FF) + (Opcode3 Shl 8)) +
                       ',R' + IntToStr(Opcode And 7);
            Inc(Addr,2);
          End
          Else
          Begin
            St := 'MVI    ' + HexWord(Opcode2) +
                       ',R' + IntToStr(Opcode And 7);
            Inc(Addr);
          End;
        End;
648..703: Begin
          Case (Opcode Shr 3) And 7 Of
            0..5: St := 'MVI@   R' + IntToStr((Opcode Shr 3) And 7) +
                              ',R' + IntToStr(Opcode And 7);
               6: St := 'PULR   R' + IntToStr(Opcode And 7);
               7: Begin
                    If SDBD Then
                    Begin
                      St := 'MVII   ' + HexWord((Opcode2 And $FF) + (Opcode3 Shl 8)) +
                                 ',R' + IntToStr(Opcode And 7);
                      Inc(Addr,2);
                    End
                    Else
                    Begin
                      St := 'MVII   ' + HexWord(Opcode2) +
                                 ',R' + IntToStr(Opcode And 7);
                      Inc(Addr);
                    End;
                  End;
          End; { Case }
        End;
 704..767: Disasm_ALU_Mem(704,'ADD ','ADD@','ADDI');
 768..831: Disasm_ALU_Mem(768,'SUB ','SUB@','SUBT');
 832..895: Disasm_ALU_Mem(832,'CMP ','CMP@','CMPI');
 896..959: Disasm_ALU_Mem(896,'AND ','AND@','ANDI');
960..1023: Disasm_ALU_Mem(960,'XOR ','XOR@','XORI');
  End; { Case }
  Inc(Addr);
  Disassemble := St;
End; { Disassemble }


Procedure Go_Mode(Mode: Byte);
Var Regs: Registers;
Begin
  Regs.AH := 0;
  Regs.AL := Mode;
  Intr($10,Regs);
End; { Go_Mode }


Const
  VGA_SEGMENT   = $A000;
  SC_INDEX      = $3C4;   { Sequence Controller Index register   }
  GC_INDEX      = $3CE;   { Graphics Controller Index register   }
  CRTC_INDEX    = $3D4;   { CRT Controller Index register        }
  MISC_OUTPUT   = $3C2;   { Miscellaneous Output Register        }
  MISC_INPUT    = $3CC;   { Miscellaneous Input Register         }
  MISCELLANEOUS = 6;      { Moscellaneous register index in GC   }
  MEMORY_MODE   = 4;      { Memory Mode register index in SC     }
  GRAPHICS_MODE = 5;      { Graphics Mode register index in GC   }
  MAX_SCAN_LINE = 9;      { Maximum Scan Line reg index in CRTC  }
  UNDERLINE     = $14;    { Underline Location reg index in CRTC }
  MODE_CONTROL  = $17;    { Mode Control register index in CRTC  }
  MAP_MASK      = 2;      { Map Mask register index in SC        }

  HORZ_TOTAL         = 0;
  HORZ_DISPLAY_END   = 1;
  START_HORZ_BLANK   = 2;
  END_HORZ_BLANK     = 3;
  START_HORZ_RETRACE = 4;
  END_HORZ_RETRACE   = 5;
  VERT_TOTAL         = 6;
  OVERFLOW           = 7;
  START_VERT_RETRACE = $10;
  END_VERT_RETRACE   = $11;
  VERT_DISPLAY_END   = $12;
  CRTC_OFFSET        = $13;
  START_VERT_BLANK   = $15;
  END_VERT_BLANK     = $16;
  LINE_COMPARE       = $18;

Procedure Set320By200Mode; Assembler;
Asm
  MOV   DX,SC_INDEX
  MOV   AL,MEMORY_MODE
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  OR    AL,08h                    { Turn on chain 4 }
  AND   AL,NOT 04h                { Turn on odd/even }
  OUT   DX,AL

  MOV   DX,GC_INDEX
  MOV   AL,GRAPHICS_MODE
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  OR    AL,10h                    { Turn on odd/even }
  OUT   DX,AL
  DEC   DX

  MOV   AL,MISCELLANEOUS
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  OR    AL,02h                    { Turn on chain }
  OUT   DX,AL

  { Now clear the whole screen.  Do this before we switch the CRTC out }
  { of mode 13h, so we don't see garbage on the screen when we make    }
  { the switch.                                                        }

  MOV   DX,SC_INDEX               { Enable writes to all planes, so  }
  MOV   AX,0F00h + MAP_MASK       {  we can clear 4 pixels at a time }
  OUT   DX,AX
  MOV   AX,VGA_SEGMENT
  MOV   ES,AX
  SUB   DI,DI
  MOV   AX,DI
  MOV   CX,8000h                  { # of words in 64k }
  CLD
  REP   STOSW

  MOV   DX,CRTC_INDEX
  MOV   AL,MAX_SCAN_LINE
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  MOV   BL,AL
  AND   BL,0E0h
  SHL   AL,1
  AND   AL,1Fh
  OR    AL,1                      { Set maximum scan line }
  OR    AL,BL
  OUT   DX,AL
  DEC   DX

  MOV   AL,UNDERLINE
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  OR    AL,40h                    { Turn on doubleword }
  OUT   DX,AL
  DEC   DX

  MOV   AL,MODE_CONTROL
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  AND   AL,NOT 40h
  OUT   DX,AL
End; { Set320By200Mode }


Procedure Set320By400Mode; Assembler;
Asm

  { Change CPU addressing of video memory to linear (not odd/even,     }
  { chain, or chain 4), to allow us to access all 256K of display      }
  { memory.  When this is done, VGA memory will look just like memory  }
  { in modes 10h and 12h, except that each byte of display memory will }
  { control one 256-color pixel, with 4 adjacent pixels at any given   }
  { address, one pixel per plane.                                      }

  MOV   DX,SC_INDEX
  MOV   AL,MEMORY_MODE
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  AND   AL,NOT 08h                { Turn off chain 4 }
  OR    AL,04h                    { Turn off odd/even }
  OUT   DX,AL

  MOV   DX,GC_INDEX
  MOV   AL,GRAPHICS_MODE
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  AND   AL,NOT 10h                { Turn off odd/even }
  OUT   DX,AL
  DEC   DX

  MOV   AL,MISCELLANEOUS
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  AND   AL,NOT 02h                { Turn off chain }
  OUT   DX,AL

  { Now clear the whole screen.  Do this before we switch the CRTC out }
  { of mode 13h, so we don't see garbage on the screen when we make    }
  { the switch.                                                        }

  MOV   DX,SC_INDEX               { Enable writes to all planes, so  }
  MOV   AX,0F00h + MAP_MASK       {  we can clear 4 pixels at a time }
  OUT   DX,AX
  MOV   AX,VGA_SEGMENT
  MOV   ES,AX
  SUB   DI,DI
  MOV   AX,DI
  MOV   CX,8000h                  { # of words in 64k }
  CLD
  REP   STOSW

  { Tweak the mode to 320x400 256-color mode by not scanning each line }
  { twice.                                                             }

  MOV   DX,CRTC_INDEX
  MOV   AL,MAX_SCAN_LINE
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  MOV   BL,AL
  AND   BL,0E0h                     { Save high three bits }
  SHR   AL,1
  AND   AL,0Fh
  OR    AL,BL
{  AND   AL,NOT 1Fh}                { Set maximum scan line = 0 }
  OUT   DX,AL
  DEC   DX

  { Change CRTC scanning from doubleword mode to byte mode, allowing }
  { the CRTC to scan more than 64K of video data.                    }

  MOV   AL,UNDERLINE
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  AND   AL,NOT 40h                { Turn off doubleword }
  OUT   DX,AL
  DEC   DX

  MOV   AL,MODE_CONTROL
  OUT   DX,AL
  INC   DX
  IN    AL,DX
  OR    AL,40h                    { Turn on the byte mode bit, so memory   }
  OUT   DX,AL                     {  is scanned for video data in a purely }
                                  {  linear way, just as in modes 10h and  }
                                  {  12h                                   }
End; { Set320By400Mode }


Procedure PutChar(X,Y: Word; C: Char; Color: Word); Assembler;
Asm
  TEST  BYTE PTR HiRes,0FFh
  JZ    @Normal
  MOV   AX,WORD PTR Y
  SHL   AX,4
  MOV   DI,AX
  SHL   DI,2
  ADD   DI,AX
  MOV   AX,WORD PTR X
  SHR   AX,2
  ADD   DI,AX
  MOV   AH,11h
  MOV   CX,WORD PTR X
  AND   CX,3
  ROL   AH,CL
  MOV   AL,MAP_MASK
  MOV   DX,SC_INDEX

  PUSH  DS
  LDS   SI,DWORD PTR FontPtr
  MOV   BL,BYTE PTR C
  SUB   BH,BH
  SHL   BX,3
  ADD   SI,BX
  MOV   BX,0A000h
  MOV   ES,BX
  MOV   CX,8
  MOV   BX,WORD PTR Color
@Loop1:
  OUT   DX,AX

  TEST  BYTE PTR [SI],80h
  JNZ   @On0
  MOV   BYTE PTR [ES:DI],BH
  JMP   @Do4
@On0:
  MOV   BYTE PTR [ES:DI],BL
@Do4:
  TEST  BYTE PTR [SI],8
  JNZ   @On4
  MOV   BYTE PTR [ES:DI + 1],BH
  JMP   @Do1
@On4:
  MOV   BYTE PTR [ES:DI + 1],BL
@Do1:

  ROL   AH,1
  ADC   DI,0
  OUT   DX,AX
  TEST  BYTE PTR [SI],40h
  JNZ   @On1
  MOV   BYTE PTR [ES:DI],BH
  JMP   @Do5
@On1:
  MOV   BYTE PTR [ES:DI],BL
@Do5:
  TEST  BYTE PTR [SI],4
  JNZ   @On5
  MOV   BYTE PTR [ES:DI + 1],BH
  JMP   @Do2
@On5:
  MOV   BYTE PTR [ES:DI + 1],BL
@Do2:

  ROL   AH,1
  ADC   DI,0
  OUT   DX,AX
  TEST  BYTE PTR [SI],20h
  JNZ   @On2
  MOV   BYTE PTR [ES:DI],BH
  JMP   @Do6
@On2:
  MOV   BYTE PTR [ES:DI],BL
@Do6:
  TEST  BYTE PTR [SI],2
  JNZ   @On6
  MOV   BYTE PTR [ES:DI + 1],BH
  JMP   @Do3
@On6:
  MOV   BYTE PTR [ES:DI + 1],BL
@Do3:

  ROL   AH,1
  ADC   DI,0
  OUT   DX,AX
  TEST  BYTE PTR [SI],10h
  JNZ   @On3
  MOV   BYTE PTR [ES:DI],BH
  JMP   @Do7
@On3:
  MOV   BYTE PTR [ES:DI],BL
@Do7:
  TEST  BYTE PTR [SI],1
  JNZ   @On7
  MOV   BYTE PTR [ES:DI + 1],BH
  JMP   @Done
@On7:
  MOV   BYTE PTR [ES:DI + 1],BL
@Done:
  ROL   AH,1
  TEST  BYTE PTR X,3
  JZ    @NoDec
  DEC   DI
@NoDec:
  ADD   DI,80
  INC   SI
  DEC   CX
  JNZ   @Loop1

  POP   DS
  JMP   @L3

@Normal:
  MOV   AX,WORD PTR Y
  SUB   BX,BX
  MOV   BH,AL
  SHL   AX,6
  ADD   BX,AX
  MOV   DI,BX
  ADD   DI,WORD PTR X
  PUSH  DS
  LDS   SI,DWORD PTR FontPtr
  MOV   AL,BYTE PTR C
  SUB   AH,AH
  SHL   AX,3
  ADD   SI,AX
  MOV   AX,0A000h
  MOV   ES,AX
  MOV   AL,8
  ADD   DI,7
  MOV   DX,WORD PTR Color
@Loop:
  MOV   BH,BYTE PTR [SI]
  MOV   CX,8
  MOV   BL,1
@L1:
  TEST  BH,BL
  JZ    @L2
  MOV   BYTE PTR [ES:DI],DL
  DEC   DI
  ADD   BL,BL
  LOOP  @L1
  ADD   DI,328
  INC   SI
  DEC   AL
  JNZ   @Loop
  POP   DS
  JMP   @L3
@L2:
  MOV   BYTE PTR [ES:DI],DH
  DEC   DI
  ADD   BL,BL
  LOOP  @L1
  ADD   DI,328
  INC   SI
  DEC   AL
  JNZ   @Loop
  POP   DS
@L3:
End; { PutChar }


Procedure WriteOut(X,Y: Word; St: String; Color: Word);
Var I: Byte;
Begin
  If St = '' Then Exit;
  For I := 0 To Length(St) - 1 Do PutChar(X + (I * 8),Y,St[I + 1],Color);
End; { WriteOut }


Procedure Swap(Var A,B: Word);
Var C: Word;
Begin
  C := A;
  A := B;
  B := C;
End; { Swap }


Procedure SwapInt(Var A,B: Integer);
Var C: Integer;
Begin
  C := A;
  A := B;
  B := C;
End; { SwapInt }


Function Debug(Buf: CP1600Mem; Var Regs: CP1600Regs; Var Flags: Word;
               Var HaltLine: Integer; IOWriteProc,IOReadProc: TIOProc;
               Const SpriteX,SpriteY: TSpriteLoc;
               PaintProc: TPaintProc): Boolean;
Var
  I,J,K    : Word;
  B        : Byte;
  Opcode   : Byte;
  Addr     : Word;
  Color    : Word;
  TopAddr  : Word;
  Ch,Ch1   : Char;
  Access   : Word;
  AccSize  : Word;
  Switch   : Boolean;
  Temp     : Boolean;
  _DLines  : Word;
  St       : String;
  HaltNum  : Integer;
  Code     : Integer;
  HexVal   : LongInt;
  HexAddr  : LongInt;
  OutFile  : File;
  SDBD     : Boolean;
  SDBDInst : Array[0..HRDisasmLines - 1] Of Boolean;
  AsmWord  : Array[0..2] Of Word;
  NumWords : Word;
  Warn     : Boolean;
  App      : PVGAApplication;
  Dir      : DirStr;
  Name     : NameStr;
  Ext      : ExtStr;
  OldSW    : Integer;
  OldSH    : Integer;
  DOSRegs  : Registers;

  Function TransWordWrite(Addr: Word): String;
  Begin
    TransWordWrite := HexWord(Addr);
(*
    If (Addr < $80) Or ((Addr >= $100) And (Addr < $180)) Then Addr := Addr And $3F;
    Case Addr Of
      0..$2C: TransWordWrite := TIA_Name[Addr];
    $200..$3FF: TransWordWrite := PIA_Name[$200 Or Timer_And[Addr And $1F]];
    Else TransWordWrite := HexWord(Addr);
    End; { Case }
*)
  End; { TransWordWrite }


  Function TransWordRead(Addr: Word): String;
  Begin
    TransWordRead := HexWord(Addr);
(*
    Case Addr Of
      0..$7F,$100..$17F: TransWordRead := TIA_Read_Name[Addr And $F];
    $200..$3FF: TransWordRead := PIA_Name[$200 Or Timer_And[Addr And $1F]];
     Else TransWordRead := HexWord(Addr);
     End; { Case }
*)
  End; { TransWordRead }


  Function TransWord(Addr: Word): String;
  Var Instr: String[3];
  Begin
    TransWord := HexWord(Addr);
(*
    Instr := Mnem(Opcode);
    If (Addr < $80) And (Copy(Instr,1,2) = 'RO')
     Then TransWord := TransWordRead(Addr)
     Else
      If (Instr           = 'ADC') Or
         (Instr           = 'DEC') Or
         (Instr           = 'INC') Or
         (Instr           = 'SBC') Or
         (Copy(Instr,1,2) = 'PH')  Or
         (Copy(Instr,1,2) = 'RO')  Or
         (Copy(Instr,1,2) = 'ST')
       Then TransWord := TransWordWrite(Addr)
       Else TransWord := TransWordRead(Addr);
*)
  End; { TransWord }


  Procedure SaveGameScreen;
  Var P: BPtr;
  Begin
    GetMem(Win,GameScreenSize);
    P := Win;
    Asm
      PUSH  DS
      LES   DI,DWORD PTR P
      SUB   SI,SI
      MOV   AX,0A000h
      MOV   DS,AX
      CLD
      MOV   DX,204
@L2:
      MOV   CX,160
@L1:
      MOVSB
      INC   SI
      LOOP  @L1
      DEC   DX
      JNZ   @L2
      POP   DS
    End; { Asm }
  End; { SaveGameScreen }


  Procedure RestoreRow(Row: Word);
  Var
    Addr : Word;
    P    : BPtr;

  Begin
    Addr := Row * 160;
    P    := Win;
    Inc(LongInt(P),Addr);
    Asm
      PUSH  DS
      MOV   DI,WORD PTR Addr
      ADD   DI,DI
      MOV   AX,0A000h
      MOV   ES,AX
      LDS   SI,DWORD PTR P
      MOV   CX,160
      CLD
@L1:
      LODSB
      MOV   AH,AL
      STOSW
      LOOP  @L1
      POP   DS
    End; { Asm }
  End; { RestoreRow }


  Procedure RestoreColumn(Col: Word);
  Var P: BPtr;
  Begin
    P := Win;
    Asm
      PUSH  DS
      MOV   DI,WORD PTR Col
      ADD   DI,DI
      MOV   AX,0A000h
      MOV   ES,AX
      LDS   SI,DWORD PTR P
      ADD   SI,WORD PTR Col
      MOV   CX,204
      CLD
@L1:
      LODSB
      MOV   AH,AL
      STOSW
      ADD   SI,159
      ADD   DI,318
      LOOP  @L1
      POP   DS
    End; { Asm }
  End; { RestoreColumn }


  Procedure RestoreGameScreen;
  Var I: Integer;
  Begin
    For I := 0 To 203 Do RestoreRow(I);
    FreeMem(Win,GameScreenSize);
    Win := Nil;
  End; { RestoreGameScreen }


  Procedure FlipWindow;
  Var
    I,J       : Word;
    Pal       : Array[0..767] Of Byte;
    Buf       : Array[0..319] Of Byte;
    MarkObj   : Boolean;
    Cnt       : Integer;
    Pos       : Integer;
    Ch        : Char;

  Begin
    Port[$3C7] := 0;
    For I := 0 To 767 Do Pal[I] := Port[$3C9];
    Temp := HiRes;
    If HiRes Then Set320By200Mode;
    HiRes := False;
    For I := 0 To 203 Do RestoreRow(I);
    MarkObj := False;
    Repeat
      Ch := UpCase(ReadKey);
      Case Ch Of
        '-': Begin
               MarkObj := Not MarkObj;
               If MarkObj Then
               Begin
                 For I := 0 To 7 Do
                 Begin
                   WriteOut((SpriteX[I] + HDelay) * 2,
                            SpriteY[I] + VertOffset + VDelay * 2,
                            IntToStr(I + 1),MarkTextColor + (MarkBackColor Shl 8));
                 End; { For I }
               End
               Else For I := 0 To 203 Do RestoreRow(I);
             End;
      End; { Case }
    Until Not (Ch In ['-']);
    HiRes := Temp;
    If HiRes Then Set320By400Mode;
    Port[$3C8] := 0;
    For I := 0 To 767 Do Port[$3C9] := Pal[I];
  End; { FlipWindow }


  Procedure Display;
  Var
    I,J,K : Integer;
    X     : Integer;
    St    : String;

    Procedure ShowBasicInfo;
    Var I: Integer;
    Begin
      { Show Help string }

      WriteOut(254,0,'F1-Help',FlagsColor + (BackColor Shl 8));

      Box(252,8,319,18,LineColor);
      Box(252,20,319,86,LineColor);
      Box(252,88,319,131,LineColor);
      FBox(253,9,318,17,FlagsColor);
      FBox(253,21,318,85,RegsColor);
      FBox(253,89,318,130,ScrDatColor);

      WriteOut(254,10,'SZOC',NormTextColor + (FlagsColor Shl 8));
      WriteOut(254,22,'R0=    h',NormTextColor + (RegsColor Shl 8));
      WriteOut(254,30,'R1=    h',NormTextColor + (RegsColor Shl 8));
      WriteOut(254,38,'R2=    h',NormTextColor + (RegsColor Shl 8));
      WriteOut(254,46,'R3=    h',NormTextColor + (RegsColor Shl 8));
      WriteOut(254,54,'R4=    h',NormTextColor + (RegsColor Shl 8));
      WriteOut(254,62,'R5=    h',NormTextColor + (RegsColor Shl 8));
      WriteOut(254,70,'R6=    h',NormTextColor + (RegsColor Shl 8));
      WriteOut(254,78,'R7=    h',NormTextColor + (RegsColor Shl 8));
{      WriteSmOut(254, 91,'Bank=   ',NormTextColor + (ScrDatColor Shl 8));}

      { Display flag data }

      For I := 0 To 3 Do
      Begin
        If (Flags And (1 Shl I)) <> 0
         Then WriteOut(310 - I * 8,10,'1',HiTextColor + (FlagsColor Shl 8))
         Else WriteOut(310 - I * 8,10,'0',HiTextColor + (FlagsColor Shl 8));
      End; { For I }

      { Display reg data }

      WriteOut(278,22,HexWord(Regs[0]),HiTextColor + (RegsColor Shl 8));
      WriteOut(278,30,HexWord(Regs[1]),HiTextColor + (RegsColor Shl 8));
      WriteOut(278,38,HexWord(Regs[2]),HiTextColor + (RegsColor Shl 8));
      WriteOut(278,46,HexWord(Regs[3]),HiTextColor + (RegsColor Shl 8));
      WriteOut(278,54,HexWord(Regs[4]),HiTextColor + (RegsColor Shl 8));
      WriteOut(278,62,HexWord(Regs[5]),HiTextColor + (RegsColor Shl 8));
      WriteOut(278,70,HexWord(Regs[6]),HiTextColor + (RegsColor Shl 8));
      WriteOut(278,78,HexWord(Regs[7]),HiTextColor + (RegsColor Shl 8));

      { Display screen info data }

{      WriteOut(284, 92,IntToStr(Game.CurrentBank)  ,HiTextColor + (ScrDatColor Shl 8));}

      { Display screen mode information }

      WriteSmOut(254,90,'Mode:',NormTextColor + (ScrDatColor Shl 8));
      If ScreenMode = smForeBack
       Then WriteSmOut(284,90,'F/B',HiTextColor + (ScrDatColor Shl 8))
       Else WriteSmOut(284,90,'ClStk',HiTextColor + (ScrDatColor Shl 8));

      { Show color stack information }

      WriteSmOut(254,101,'CStk+BdCol',NormTextColor + (ScrDatColor Shl 8));
      For I := 0 To 3 Do
      Begin
        J  := Buf[0]^[ColorStackLoc + I] And $F;
        St := IntToStr(J);
        K  := Palette[J].R + Palette[J].G + Palette[J].B;
        X  := 254 + I * 12;
        FBox(X,109,X + 11,117,J);
        If Length(St) = 1 Then Inc(X,3);
        If K >= 32*3
         Then WriteSmOut(X,110,St,J Shl 8)
         Else WriteSmOut(X,110,St,(J Shl 8) + 7);
      End; { For I }

      { Show border information }

      J  := Buf[0]^[BorderColorLoc] And $F;
      St := IntToStr(J);
      K  := Palette[J].R + Palette[J].G + Palette[J].B;
      X  := 305;
      FBox(X,109,X + 11,117,J);
      If Length(St) = 1 Then Inc(X,3);
      If K >= 32*3
       Then WriteSmOut(X,110,St,J Shl 8)
       Else WriteSmOut(X,110,St,(J Shl 8) + 7);

      { Show horizontal & vertical delay }

      WriteSmOut(254,122,'Delay=',NormTextColor + (ScrDatColor Shl 8));
      WriteSmOut(290,122,IntToStr(HDelay),HiTextColor + (ScrDatColor Shl 8));
      WriteSmOut(296,122,',',NormTextColor + (ScrDatColor Shl 8));
      WriteSmOut(302,122,IntToStr(VDelay),HiTextColor + (ScrDatColor Shl 8));
    End; { ShowBasicInfo }

    Procedure ShowDataArea(Y: Integer);
    Var I,J: Integer;
    Begin
      For I := 0 To DataLines - 1 Do
      Begin
        WriteOut(2,Y + I * 8,HexWord((DataTop + I * 8) And $3FF),DataTextColor + (DataColor Shl 8));
        For J := 0 To 7 Do
        Begin
          WriteSmOut(44 + J * 29,Y + I * 8,
           HexWord(Buf[((DataTop + I * 8 + J) And $3FF) Shr 12]^[((DataTop + I * 8 + J) And $3FF) And $FFF]),
           DataTextColor + (DataColor Shl 8));
        End; { For J }
      End; { For I }
    End; { ShowDataArea }

  Begin
    If HiRes Then
    Begin
      FBox(0,0,319,399,BackColor);
      Box(0,0,250,259,LineColor);
      Box(0,333,319,399,LineColor);
      Box(0,261,319,287,LineColor);
      Box(0,289,319,299,LineColor);
      Box(0,301,319,320,LineColor);
      FBox(1,1,249,258,DisasmColor);
      FBox(1,334,318,398,DataColor);
      FBox(1,262,318,286,TIA_Color);
      FBox(1,290,318,298,TIA_Color);
      FBox(1,302,318,319,DataColor);

      ShowBasicInfo;

      { Display object locations }

      For I := 0 To 3 Do
      Begin
        J  := (Buf[0]^[$10 + I] And $7) + ((Buf[0]^[$10 + I] Shr 9) And $8);
        K  := Palette[J].R + Palette[J].G + Palette[J].B;
        J  := J Shl 8;
        If K < 32*3 Then Inc(J,7);
        St := 'S' + IntToStr(I + 1) + '=';
        WriteOut(2 + I * 80,303,St,J);
        St := '(' + IntToStr(SpriteX[I]) + ',' + IntToStr(SpriteY[I]) + ')';
        WriteSmOut(18 + I * 80,303,St,ObjLocColor + (DataColor Shl 8));
      End; { For I }
      For I := 4 To 7 Do
      Begin
        J  := (Buf[0]^[$10 + I] And $7) + ((Buf[0]^[$10 + I] Shr 9) And $8);
        K  := Palette[J].R + Palette[J].G + Palette[J].B;
        J  := J Shl 8;
        If K < 32*3 Then Inc(J,7);
        St := 'S' + IntToStr(I + 1) + '=';
        WriteOut(2 + (I - 4) * 80,311,St,J);
        St := '(' + IntToStr(SpriteX[I]) + ',' + IntToStr(SpriteY[I]) + ')';
        WriteSmOut(18 + (I - 4) * 80,311,St,ObjLocColor + (DataColor Shl 8));
      End; { For I }

      { Display data area }

      ShowDataArea(335);
    End
    Else
    Begin
      FBox(0,0,319,199,BackColor);
      Box(0,0,250,131,LineColor);
      Box(0,133,319,199,LineColor);
      FBox(1,1,249,130,DisasmColor);
      FBox(1,134,318,198,DataColor);

      ShowBasicInfo;

      { Display data area }

      ShowDataArea(135);
    End;
  End; { Display }


  Procedure GetImage(X1,Y1,X2,Y2: LongInt; Var P);
  Var X,Y,Addr,Dest: LongInt;
  Begin
    If HiRes Then
    Begin
      For Y := Y1 To Y2 Do
      Begin
        Addr := Y * 320 + X1;
        Dest := (Y - Y1) * (X2 - X1 + 1);
        For X := X1 To X2 Do
        Begin
          Asm
            CLI
            MOV   DX,03CEh
            MOV   AL,4
            MOV   AH,BYTE PTR Addr
            ADD   AH,BYTE PTR X
            SUB   AH,BYTE PTR X1
            AND   AH,3
            OUT   DX,AX
            STI
          End; { Asm }
          Mem[Seg(P):Ofs(P) + Dest + (X - X1)] := Mem[$A000:(Addr + X - X1) Shr 2];
        End; { For X }
      End; { For Y }
      Asm
        CLI
        MOV   DX,03CEh
        MOV   AL,4
        MOV   AH,0
        OUT   DX,AX
        STI
      End; { Asm }
    End
    Else
    Begin
      For Y := Y1 To Y2 Do
      Begin
        Addr := Y * 320 + X1;
        Dest := (Y - Y1) * (X2 - X1 + 1);
        Move(Mem[$A000:Addr],Mem[Seg(P):Ofs(P) + Dest],X2 - X1 + 1);
      End; { For Y }
    End;
  End; { GetImage }


  Procedure PutImage(X1,Y1,W,H: Word; Var P);
  Var X,Y,Addr,Source: LongInt;
  Begin
    If HiRes Then CopySystemToScreenX(0,0,W,H,X1,Y1,@P,0,W,320)
    Else
    Begin
      For Y := Y1 To Y1 + H - 1 Do
      Begin
        Addr   := Y * 320 + X1;
        Source := (Y - Y1) * W;
        Move(Mem[Seg(P):Ofs(P) + Source],Mem[$A000:Addr],W);
      End; { For Y }
    End;
  End; { PutImage }


  Function GetText(X1,Y1,X2,Y2: Integer; InitString: String; Fore,Back: Byte; MaxLen,WindowLen: Integer): String;
  { ---------------------------------------------------------------------- }
  { Allows the user to type in a text string in graphics mode              }
  {    X1 ................. Upper left X coord of text input box           }
  {    Y1 ................. Upper left Y coord of text input box           }
  {    X2 ................. Lower right X coord of text input box          }
  {    Y2 ................. Lower right Y coord of text input box          }
  {    InitString ......... Initial text string                            }
  {    Fore ............... Text foreground color                          }
  {    Back ............... Text background color                          }
  {    MaxLen ............. Maximum string length                          }
  {    WindowLen .......... Displayed string length, for scrollable window }
  { ---------------------------------------------------------------------- }
  Const
    TextWidth  = 8;
    TextHeight = 8;

  Var
    Work  : String;
    Ch    : Char;
    Ch1   : Char;
    Top   : Integer;
    X     : Integer;
    Ins   : Boolean;
    TPtr  : Pointer;
    On    : Boolean;
    Pen   : Word;
    Brush : Word;

    Function ImageSize(X1,Y1,X2,Y2: Word): Word;
    Begin
      ImageSize := (X2 - X1 + 1) * (Y2 - Y1 + 1);
    End; { ImageSize }


    Function Max(A,B: Integer): Integer;
    Begin
      If A >= B Then Max := A Else Max := B;
    End; { Max }


    Function Min(A,B: Integer): Integer;
    Begin
      If A <= B Then Min := A Else Min := B;
    End; { Min }


    Procedure Cursor;
    { --------------------------------- }
    { Toggle to display/hide the cursor }
    { --------------------------------- }
    Begin
      If On Then
      Begin
        Pen   := Back;
        Brush := Back;
      End
      Else
      Begin
        Pen   := Fore;
        Brush := Fore;
      End;
      If Ins
       Then FBox(X1 + (X - Top)     * TextWidth + 1,Y1 + 4,
                 X1 + (X - Top + 1) * TextWidth,Y1 + TextHeight,Brush)
       Else FBox(X1 + (X - Top)     * TextWidth + 1,Y1 + TextHeight,
                 X1 + (X - Top + 1) * TextWidth,Y1 + TextHeight,Pen);
      On := Not On;
      If Not On Then
      Begin
        Pen   := Fore;
        Brush := Back;
        WriteOut(X1 + (X - Top) * TextWidth + 1,Y1 + 1,
         Copy(Work,X,1),Pen + (Brush Shl 8));
      End;
      Pen   := Fore;
      Brush := Fore;
    End; { Cursor }


    Procedure ScrollLeft;
    { --------------------------- }
    { Scroll the text window left }
    { --------------------------- }
    Var W,H,S: Word;
    Begin
      Inc(Top);
      S := ImageSize(X1 + TextWidth + 1,Y1,X2,Y2);
      GetMem(TPtr,S);
      GetImage(X1 + TextWidth + 1,Y1,X2,Y2,TPtr^);
      W     := X2 - (X1 + TextWidth + 1) + 1;
      H     := Y2 - Y1 + 1;
      Pen   := Back;
      Brush := Back;
      FBox(X1 + (WindowLen - 1) * TextWidth + 1,Y1,X2,Y2,Brush);
      Pen   := Fore;
      PutImage(X1 + 1,Y1,W,H,TPtr^);
      WriteOut(X1 + (WindowLen) * TextWidth + 1,Y1 + 1,
       Copy(Work,Top + WindowLen,1),Pen + (Brush Shl 8));
      FreeMem(TPtr,S);
    End; { ScrollLeft }


    Procedure ScrollRight;
    { ---------------------------- }
    { Scroll the text window right }
    { ---------------------------- }
    Var W,H,S: Word;
    Begin
      Dec(Top);
      S := ImageSize(X1 + TextWidth + 1,Y1,X2 + (WindowLen - 1) * TextWidth,Y2);
      GetMem(TPtr,S);
      GetImage(X1 + 1,Y1,X2 - TextWidth,Y2,TPtr^);
      W     := (X2 - TextWidth) - (X1 + 1) + 1;
      H     := Y2 - Y1 + 1;
      Pen   := Back;
      Brush := Back;
      FBox(X1,Y1,X1 + TextWidth,Y2,Brush);
      Pen   := Fore;
      PutImage(X1 + TextWidth + 1,Y1,W,H,TPtr^);
      WriteOut(X1 + 1,Y1 + 1,Copy(Work,Top,1),Pen + (Brush Shl 8));
      FreeMem(TPtr,S);
    End; { ScrollRight }

  Begin
    Work := InitString;{'';}
    Top  := 1;
    X    := 1;
    Ins  := False;
    On   := False;

    { Blank the window }

    Pen   := Back;
    Brush := Back;
    FBox(X1,Y1,X2,Y2,Brush);
    Pen   := Fore;

    WriteOut(X1 + 1,Y1 + 1,Copy(Work,1,WindowLen + 1),Pen + (Brush Shl 8));

    Cursor;
    Ch := ' ';
    While Not (Ch In [#13,#27]) Do
    Begin
      If KeyPressed Then
      Begin
        Ch := ReadKey;
        Case Ch Of
     #13,#27: Begin
              End;
          #8: Begin                                   { Backspace }
                If X > 1 Then
                Begin
                  Cursor;
                  Delete(Work,X - 1,1);
                  If X > Top Then
                  Begin
                    Pen   := Back;
                    Brush := Back;
                    FBox(X1 + (X - Top - 1) * TextWidth + 1,Y1 + 1,X2 - 1,Y2 - 1,Brush);
                    Pen   := Fore;
                    WriteOut(X1 + (X - Top - 1) * TextWidth + 1,Y1 + 1,
                             Copy(Work,X - 1,WindowLen - (X - Top) + 1),Pen + (Brush Shl 8));
                  End;
                  Dec(X);
                  If X < Top Then ScrollRight;
                  Cursor;
                End;
              End;
          #0: Begin
                While Not KeyPressed Do Begin End;
                Ch1 := ReadKey;
                Case Ch1 Of
                  #71: If X > 1 Then                  { Home }
                       Begin
                         Cursor;
                         If Top <> 1 Then
                         Begin
                           Pen   := Back;
                           Brush := Back;
                           FBox(X1,Y1,X2,Y2,Brush);
                           Pen := Fore;
                           Top := 1;
                           WriteOut(X1 + 1,Y1 + 1,Copy(Work,1,WindowLen + 1),Pen + (Brush Shl 8));
                         End;
                         X := 1;
                         Cursor;
                       End;
                  #75: If X > 1 Then                  { Cursor Left }
                       Begin
                         Cursor;
                         Dec(X);
                         If X < Top Then ScrollRight;
                         Cursor;
                       End;
                  #77: If X <= Length(Work) Then      { Cursor Right }
                       Begin
                         Cursor;
                         Inc(X);
                         If X - Top = WindowLen Then ScrollLeft;
                         Cursor;
                       End;
                  #79: If X <= Length(Work) Then      { End }
                       Begin
                         Cursor;
                         If Top < Length(Work) - WindowLen + 2 Then
                         Begin
                           Pen   := Back;
                           Brush := Back;
                           FBox(X1,Y1,X2,Y2,Brush);
                           Pen   := Fore;
                           Top   := Length(Work) - WindowLen + 2;
                           WriteOut(X1 + 1,Y1 + 1,Copy(Work,Top,WindowLen),Pen + (Brush Shl 8));
                         End;
                         X := Length(Work) + 1;
                         Cursor;
                       End;
                  #82: Begin                          { Insert }
                         Cursor;
                         Ins := Not Ins;
                         Cursor;
                       End;
                  #83: Begin                          { Delete }
                         If ((Work <> '') And (X <= Length(Work))) Then
                         Begin
                           Cursor;
                           Delete(Work,X,1);
                           Pen   := Back;
                           Brush := Back;
                           FBox(X1 + (X - Top) * TextWidth + 1,Y1 + 1,X2 - 1,Y2 - 1,Brush);
                           Pen   := Fore;
                           WriteOut(X1 + (X - Top) * TextWidth + 1,Y1 + 1,
                            Copy(Work,X,WindowLen - (X - Top) + 1),Pen + (Brush Shl 8));
                           Cursor;
                         End;
                       End;
                End; { Case }
              End;
        Else
          Cursor;
          If (X > Length(Work)) Or Ins Then
          Begin
            If (Length(Work) < MaxLen) Then
            Begin
              Insert(Ch,Work,X);
              Pen   := Back;
              Brush := Back;
              FBox(X1 + (X - Top) * TextWidth + 1,Y1 + 1,X2 - 1,Y2 - 1,Brush);
              Pen   := Fore;
              WriteOut(X1 + (X - Top) * TextWidth + 1,Y1 + 1,
               Copy(Work,X,WindowLen - (X - Top) + 1),Pen + (Brush Shl 8));
            End;
          End
          Else
          Begin
            Work[X] := Ch;
            Pen   := Back;
            Brush := Back;
            FBox(X1 + (X - Top)     * TextWidth + 1,Y1 + 1,
                 X1 + (X - Top + 1) * TextWidth    ,Y2 - 1,Brush);
            Pen   := Fore;
            WriteOut(X1 + (X - Top) * TextWidth + 1,Y1 + 1,Copy(Work,X,1),Pen + (Brush Shl 8));
          End;
          If Not ((X > Length(Work)) And (Length(Work) = MaxLen)) Then
          Begin
            Inc(X);

             { Check to see if I must scroll the text left }

             If X - Top = WindowLen Then ScrollLeft;
          End;
          Cursor;
        End; { Case }
      End;
    End; { While }
    If Ch = #27 Then Work := InitString;
    Pen   := Back;
    Brush := Back;
    FBox(X1,Y1,X2,Y2,Brush);
    Pen   := Fore;
    WriteOut(X1 + 1,Y1 + 1,Copy(Work,1,WindowLen),Pen + (Brush Shl 8));
    GetText := Work;
  End; { GetText }

  Procedure DisplayHelp;
  Var
    I  : Integer;
    St : String;

  Begin
    Temp := ModeX;
    If ModeX Then Set320By200Mode;
    ModeX := False;
    HiRes := False;
    FBox(0,0,319,199,HelpColor);

    Color := HelpHiColor + (HelpColor Shl 8);
    WriteOut(0,0,  'A',Color);
    WriteOut(0,8,  'D',Color);
    WriteOut(0,16, 'E',Color);
    WriteOut(0,24, 'G',Color);
    WriteOut(0,32, 'H',Color);
    WriteOut(0,40, 'R',Color);
    WriteOut(0,48, 'S',Color);
    WriteOut(0,56, 'T',Color);
    WriteOut(0,64, 'X',Color);
    WriteOut(0,72, 'Z',Color);
    WriteOut(0,80, 'B',Color);
    WriteOut(0,88, 'L',Color);
    WriteOut(0,96, 'P',Color);
    WriteOut(0,104,'<Space>',Color);
    WriteOut(0,112,'-',Color);
    WriteOut(0,120,'[ or ]',Color);

    WriteOut(0,128,'<Up>',Color);
    WriteOut(0,136,'<Down>',Color);
    WriteOut(0,144,'<PgUp>',Color);
    WriteOut(0,152,'<PgDn>',Color);
    WriteOut(0,160,'<Home>',Color);
    WriteOut(0,168,'<End>',Color);
    WriteOut(0,176,'<Ins>',Color);

    Color := HelpTextColor + (HelpColor Shl 8);
    WriteOut(56,0,  ' - Assemble instruction',Color);
    WriteOut(56,8,  ' - Toggle 320x400 mode',Color);
    WriteOut(56,16, ' - Enter data at address',Color);
    WriteOut(56,24, ' - Begin execution of code',Color);
    WriteOut(56,32, ' - Execute to hilighted position',Color);
    WriteOut(56,40, ' - Change CP1600 register',Color);
    WriteOut(56,48, ' - Execute over next instruction',Color);
    WriteOut(56,56, ' - Execute next instruction',Color);
    WriteOut(56,64, ' - Enter break expression',Color);
    WriteOut(56,72, ' - Dump CP1600 memory to disk',Color);
    WriteOut(56,80, ' - Move to BACKTAB (200h)',Color);
    WriteOut(56,88, ' - Load save game state',Color);
    WriteOut(56,96, ' - Move to specified address',Color);
    WriteOut(56,104,' - Toggle game screen display',Color);
    WriteOut(56,112,' - (game screen) show object pos.',Color);
    WriteOut(56,120,' - Scroll data area',Color);

    WriteOut(56,128,' - Move up one byte',Color);
    WriteOut(56,136,' - Move down one instruction',Color);
    WriteOut(56,144,' - Move up one screen',Color);
    WriteOut(56,152,' - Move down one screen',Color);
    WriteOut(56,160,' - Move to 5000h',Color);
    WriteOut(56,168,' - Move to FFFFh',Color);
    WriteOut(56,176,' - Move to current instruction',Color);

    Color := ScrDatColor + (HelpColor Shl 8);
    WriteOut(0,184,'Hit <Space> for more, or',Color);
    WriteOut(0,192,'any other key to return',Color);

    Ch1 := ReadKey;
(*
    If Ch1 = #32 Then
    Begin
      FBox(0,0,319,199,HelpColor);
      Color := HelpHiColor + (HelpColor Shl 8);
      For I := 0 To $2C Do
      Begin
        St := HexByte(I);
        While Length(St) < 2 Do St := '0' + St;
        If I < $17
         Then WriteOut(8,I * 8,St,Color)
         Else WriteOut(160,(I - $17) * 8,St,Color);
      End;
      Color := HelpTextColor + (HelpColor Shl 8);
      For I := 0 To $2C Do
      Begin
        If I < $17
         Then WriteOut(32,I * 8,'- ' + TIA_Name[I],Color)
         Else WriteOut(184,(I - $17) * 8,'- ' + TIA_Name[I],Color);
      End;
      Color := ScrDatColor + (HelpColor Shl 8);
      WriteOut(0,184,'Hit <Space> for more, or',Color);
      WriteOut(0,192,'any other key to return',Color);
      Ch1 := ReadKey;
*)
      If Ch1 = #32 Then
      Begin
        FBox(0,0,319,199,HelpColor);
        Color := HelpHiColor + (HelpColor Shl 8);
        WriteOut(0,0,  ' "E" command syntax:',Color);
        WriteOut(0,48, ' "R" command syntax:',Color);
        WriteOut(0,88, ' Break expression arguments:',Color);
        WriteOut(0,160,' Assemble instruction notes:',Color);
        Color := HelpTextColor + (HelpColor Shl 8);
        WriteOut(0,8,  '  <address>=<value1> <value2> ...',Color);
        WriteOut(0,16, '  e.g. 125F=28 34D1 places 28h',Color);
        WriteOut(0,24, '       into [125Fh] and 34d1h',Color);
        WriteOut(0,32, '       into [1260h]',Color);

        WriteOut(0,56, '  <register>=<value>',Color);
        WriteOut(0,64, '  e.g. R3=28 places 28h into reg. 3',Color);
        WriteOut(0,72, '  Available registers: R0..R7, SW',Color);

        WriteOut(0,96, '  S1X..S8X .... Sprite X position',Color);
        WriteOut(0,104,'  S1Y..S8Y .... Sprite Y position',Color);
        WriteOut(0,112,'  R0..R7 ...... Register contents',Color);
        WriteOut(0,120,'  SW .......... Status word contents',Color);
        WriteOut(0,128,'  [address] ... Memory contents',Color);
        WriteOut(0,136,'  R[address] .. Read from memory',Color);
        WriteOut(0,144,'  W[address] .. Write to memory',Color);

        WriteOut(0,168,'  ALL numbers are in hexadecimal',Color);
        WriteOut(0,176,'  e.g. MVII 013F,R5',Color);

        Color := ObjLocColor + (HelpColor Shl 8);
        WriteOut(0,184,'Hit <Space> for more, or',Color);
        WriteOut(0,192,'any other key to return',Color);
        Ch1 := ReadKey;
        If Ch1 = #32 Then
        Begin
          FBox(0,0,319,199,HelpColor);
          Color := HelpHiColor + (HelpColor Shl 8);
          WriteOut(0,0,  ' Break expression operators:',Color);
          Color := HelpTextColor + (HelpColor Shl 8);
          WriteOut(0,8,  '  + ......... Addition',Color);
          WriteOut(0,16, '  - ......... Subtraction',Color);
          WriteOut(0,24, '  * ......... Multiplication',Color);
          WriteOut(0,32, '  / ......... Integer Division',Color);
          WriteOut(0,40, '  % ......... Modulo',Color);

          WriteOut(0,56, '  && ........ Boolean and',Color);
          WriteOut(0,64, '  || ........ Boolean or',Color);
          WriteOut(0,72, '  ^^ ........ Boolean xor',Color);
          WriteOut(0,80, '  ! or ~ .... Boolean not',Color);

          WriteOut(0,96, '  & ......... Bitwise and',Color);
          WriteOut(0,104,'  | ......... Bitwise or',Color);
          WriteOut(0,112,'  ^ ......... Bitwise xor',Color);

          WriteOut(0,128,'  = or == ... Bool/num equality',Color);
          WriteOut(0,136,'  <> or != .. Bool/num inequality',Color);
          WriteOut(0,144,'  < ......... Less than',Color);
          WriteOut(0,152,'  > ......... Greater than',Color);
          WriteOut(0,160,'  <= ........ Less than or equal',Color);
          WriteOut(0,168,'  >= ........ Greater than or equal',Color);

          WriteOut(0,184,'  ( or ) .... Parentheses',Color);
          ReadKey;
        End;
{      End;}
    End;

    ModeX := Temp;
    HiRes := Temp;
    If HiRes Then Set320By400Mode;
  End; { DisplayHelp }


  Function LoadSaveState: Boolean;
  Var
    St : String;
    I  : Integer;
    F  : File;
    B  : Byte;
    FB : Byte;

  Begin
    St := UpperCase(Trim(App^.FileDialog('Load save game state','*.*','')));
    If (St <> '') And Exist(St) Then
    Begin
      Assign(F,St);
      Reset(F,1);
      For I := 0 To $F Do BlockRead(F,Buf[I]^[0],$2000);

      { Load CP1600 status word }

      Seek(F,$2001D);         { Not sure about this }
      BlockRead(F,Flags,2);

      { Load CP1600 registers }

      Seek(F,$2001E);
      BlockRead(F,Regs,16);

      { Load sound information }

      Seek(F,$20008);
      For I := 0 To $F Do
      Begin
        BlockRead(F,B,1);
        IOWriteProc($1F0 + I,B);
      End; { For I }

      { Load foreground/background / color stack information }

      Seek(F,$20039);
      BlockRead(F,FB,1);

      Close(F);
      For I := 0 To $65 Do IOWriteProc(I,Buf[0]^[I]);
      If FB <> 0 Then IOWriteProc($0021,0) Else IOReadProc($0021,0);
      LoadSaveState := True;
    End
    Else LoadSaveState := False;
  End; { LoadSaveState }

Begin
  OldSW   := ScreenWidth;
  OldSH   := ScreenHeight;
  ScreenWidth := 320;
  If HiRes
   Then ScreenHeight := 400
   Else ScreenHeight := 200;
  ModeX := HiRes;
  TopAddr := Regs[7];
  SaveGameScreen;

  If HiRes Then Set320By400Mode;
  Display;

  MouseInit(ScreenWidth,ScreenHeight,2,1,ModeX);
  App := New(PVGAApplication,Init(ScreenWidth,ScreenHeight,ModeX,False,True));

  I := TopAddr - 1;
  SDBD := ((TopAddr > 0) And ((Buf[(I) Shr 12]^[(I) And $FFF] And $3FF) = 1));
  Repeat

    { Display instructions }

    Addr := TopAddr;
    If HiRes
     Then _DLines := HRDisasmLines
     Else _DLines := DisasmLines;
    For I := 0 To _DLines - 1 Do
    Begin
      If Addr = Regs[7] Then
      Begin
        Color := PCTextColor + (HiTextColor   Shl 8);
        FBox(2,2 + I * 8,248,9 + I * 8,HiTextColor);
      End
      Else
      Begin
        Color := NormTextColor + (DisasmColor Shl 8);
        FBox(2,2 + I * 8,248,9 + I * 8,DisasmColor);
      End;

      Opcode := Buf[Addr Shr 12]^[Addr And $FFF] And $3FF;
      If Opcode = 1 Then SDBD := True;

      If HiRes
       Then WriteSmOut(2,2 + I * 8,HexWord(Addr),Color)
       Else WriteOut(2,2 + I * 8,HexWord(Addr),Color);

      K := Addr;
      If HiRes
       Then WriteSmOut(124,2 + I * 8,Disassemble(Buf,Addr,SDBD),Color)
       Else WriteOut(124,2 + I * 8,Disassemble(Buf,Addr,SDBD),Color);
      SDBDInst[I] := SDBD;

      For J := 0 To Addr - K - 1 Do
      Begin
        WriteSmOut(38 + J * 28,2 + I * 8,
         HexWord(Buf[(K + J) Shr 12]^[(K + J) And $FFF]),Color);
      End; { For J }
      If Opcode <> 1 Then SDBD := False;
    End; { For I }
    Ch := UpCase(ReadKey);
    If Ch = 'D' Then
    Begin
      Dispose(App,Done);
      HiRes := Not HiRes;
      ModeX := HiRes;
      ScreenWidth := 320;
      If HiRes
       Then ScreenHeight := 400
       Else ScreenHeight := 200;
      If HiRes Then Set320By400Mode Else Set320By200Mode;
      Display;
      MScreenWidth  := ScreenWidth;
      MScreenHeight := ScreenHeight;
      UseModeX      := ModeX;

      DOSRegs.AX := 7;                           { Set min/max horizontal position }
      DOSRegs.CX := 0;
      DOSRegs.DX := (MScreenWidth * WidthMul) - 1;
      Intr(MouseInterrupt,DOSRegs);
      DOSRegs.AX := 8;                             { Set min/max vertical position }
      DOSRegs.CX := 0;
      DOSRegs.DX := (MScreenHeight * HeightMul) - 1;
      Intr(MouseInterrupt,DOSRegs);
      App := New(PVGAApplication,Init(ScreenWidth,ScreenHeight,ModeX,False,False));
    End;
    If Ch = 'G' Then DebugAddr := 0;
    If Ch = 'S' Then
    Begin
      Opcode := Buf[Regs[7] Shr 12]^[Regs[7] And $FFF] And $3FF;
      I := Regs[7];
      If (Regs[7] > 0) And ((Buf[(Regs[7] - 1) Shr 12]^[(Regs[7] - 1) And $FFF] And $3FF) = 1)
       Then Disassemble(Buf,I,True)
       Else Disassemble(Buf,I,False);
      DebugAddr := I;
    End;
    If Ch = 'H' Then DebugAddr := TopAddr;
    If Ch = 'Z' Then
    Begin
      St := UpperCase(Trim(App^.FileDialog('Save memory to disk','*.*','')));
      If St <> '' Then
      Begin
        If FileExists(St) Then
        Begin
          FSplit(St,Dir,Name,Ext);
          Temp := (App^.ShowMessage('Warning: file already exists','Overwrite file ' + Name + Ext +
                                    '?',[mbOK,mbCancel]) = idOK);
        End
        Else Temp := True;
        If Temp Then
        Begin
        {$I-}
          Assign(OutFile,St);
          ReWrite(OutFile,1);
        {$I+}
          If IOResult = 0 Then
          Begin
            For I := 0 To $F Do BlockWrite(OutFile,Buf[I]^,$2000);
            Close(OutFile);
          End;
        End;
      End;
      Display;
    End;
    If Ch = 'P' Then
    Begin
      St := Trim(App^.InputBox('Change disasm position','Enter address in hex:',''));
      If St <> '' Then
      Begin
        HexAddr := ValHex(St);
        If (HexAddr >= 0) And (HexAddr <= $FFFF) Then TopAddr := HexAddr;
      End;
      Display;
    End;
    If Ch = 'X' Then
    Begin
      St := UpperCase(Trim(App^.InputBox('Break on expression','Enter expression:',BreakExp)));
      Display;

      { Get rid of the last break expression collection, if any }

      If BreakExpColl <> Nil Then
      Begin
        Dispose(BreakExpColl,Done);
        BreakExpColl := Nil;
      End;
      ExpError := False;
      If St <> '' Then
      Begin
        BreakExpColl := PostFix(St);
        If BreakExpColl <> Nil Then
        Begin

          { Try to evaluate it to make sure it's ok (ignore the result) }

          Evaluate(BreakExpColl,Buf,Regs,Flags,SpriteX,SpriteY,False,False,0,ExpError);
          If ExpError Then
          Begin
            Dispose(BreakExpColl,Done);
            BreakExpColl := Nil;
            St := '';
          End;
        End
        Else
        Begin
          St       := '';
          ExpError := True;
        End;
      End;
      BreakExp := St;  { Remember the last one }
      If ExpError Then
      Begin
        App^.ShowMessage('Error','Error in expression',[mbOK]);
        Display;
      End;
    End;
    If Ch = 'E' Then
    Begin
      St := UpperCase(Trim(App^.InputBox('Change CP1600 memory','Enter <address=data> (hex):','')));
      If St <> '' Then
      Begin
        Code := Pos('=',St);
        If Code <> 0 Then
        Begin
          HexAddr := ValHex(Trim(Copy(St,1,Code - 1))) And $FFFF;
          St      := Trim(Copy(St,Code + 1,Length(St) - Code));
          Code    := Pos(' ',St);
          If (Code = 0) And (St <> '') Then Code := Length(St) + 1;
          While Code <> 0 Do
          Begin
            HexVal := ValHex(Trim(Copy(St,1,Code - 1)));
            If HexVal >= 0 Then
            Begin
              Buf[HexAddr Shr 12]^[HexAddr And $FFF] := HexVal And $FFFF;
              IOWriteProc(HexAddr,HexVal);
            End;
            Inc(HexAddr);
            St   := Trim(Copy(St,Code + 1,Length(St) - Code));
            Code := Pos(' ',St);
            If (Code = 0) And (St <> '') Then Code := Length(St) + 1;
          End; { While }
        End;
      End;
      Display;
    End;
    If Ch = 'A' Then
    Begin
      St := UpperCase(Trim(App^.InputBox('Assemble CP1600 instruction','Instruction:','')));
      If St <> '' Then
      Begin
        SDBD := ((TopAddr > 0) And ((Buf[(TopAddr - 1) Shr 12]^[(TopAddr - 1) And $FFF] And $3FF) = 1));
        If Assemble(St,TopAddr,SDBD,AsmWord[0],AsmWord[1],AsmWord[2],NumWords,Warn) Then
        Begin
          For I := 0 To NumWords - 1 Do
          Begin
            Buf[(TopAddr + I) Shr 12]^[(TopAddr + I) And $FFF] := AsmWord[I];
            IOWriteProc(TopAddr + I,AsmWord[I]);
          End; { For I }
          If Warn Then App^.ShowMessage('Warning','Instruction requires more than 10 bits',[mbOK]);
        End
        Else App^.ShowMessage('Error','Invalid instruction',[mbOK]);
      End;
      Display;
    End;
    If Ch = 'T' Then Trace := 1;
    If Ch = 'B' Then TopAddr := $200;
    If Ch = 'L' Then
    Begin
      If LoadSaveState Then
      Begin
        If HiRes Then Set320By200Mode;
        RestoreGameScreen;
        PaintProc;
        SaveGameScreen;
        If HiRes Then Set320By400Mode;
      End;
      Display;
    End;
    If Ch = 'R' Then
    Begin
      St := UpperCase(Trim(App^.InputBox('Change CP1600 register','Enter <reg=value> (hex):','')));
      Code := Pos(' ',St);
      While Code <> 0 Do
      Begin
        Delete(St,Code,1);
        Code := Pos(' ',St);
      End; { While }
      If St <> '' Then
      Begin
        If Copy(St,1,3) = 'R0=' Then
        Begin
          HexVal := ValHex(Copy(St,4,Length(St) - 3));
          If HexVal >= 0 Then
          Begin
            Regs[0] := HexVal;
          End;
        End;
        If Copy(St,1,3) = 'R1=' Then
        Begin
          HexVal := ValHex(Copy(St,4,Length(St) - 3));
          If HexVal >= 0 Then
          Begin
            Regs[1] := HexVal;
          End;
        End;
        If Copy(St,1,3) = 'R2=' Then
        Begin
          HexVal := ValHex(Copy(St,4,Length(St) - 3));
          If HexVal >= 0 Then
          Begin
            Regs[2] := HexVal;
          End;
        End;
        If Copy(St,1,3) = 'R3=' Then
        Begin
          HexVal := ValHex(Copy(St,4,Length(St) - 3));
          If HexVal >= 0 Then
          Begin
            Regs[3] := HexVal;
          End;
        End;
        If Copy(St,1,3) = 'R4=' Then
        Begin
          HexVal := ValHex(Copy(St,4,Length(St) - 3));
          If HexVal >= 0 Then
          Begin
            Regs[4] := HexVal;
          End;
        End;
        If Copy(St,1,3) = 'R5=' Then
        Begin
          HexVal := ValHex(Copy(St,4,Length(St) - 3));
          If HexVal >= 0 Then
          Begin
            Regs[5] := HexVal;
          End;
        End;
        If Copy(St,1,3) = 'R6=' Then
        Begin
          HexVal := ValHex(Copy(St,4,Length(St) - 3));
          If HexVal >= 0 Then
          Begin
            Regs[6] := HexVal;
          End;
        End;
        If Copy(St,1,3) = 'R7=' Then
        Begin
          HexVal := ValHex(Copy(St,4,Length(St) - 3));
          If HexVal >= 0 Then
          Begin
            Regs[7] := HexVal;
          End;
        End;
        If Copy(St,1,3) = 'SW=' Then
        Begin
          HexVal := ValHex(Copy(St,4,Length(St) - 3));
          If HexVal >= 0 Then
          Begin
            Flags := HexVal And $F;
          End;
        End;
      End;
      Display;
    End;
    If Ch = ' ' Then
    Begin
      FlipWindow;
      Display;
    End;
    If Ch = '[' Then
    Begin
      DataTop := (DataTop - 8) And $3FF;
      Display;
    End;
    If Ch = ']' Then
    Begin
      DataTop := (DataTop + 8) And $3FF;
      Display;
    End;


    If Ch = #0 Then
    Begin
      Ch := ReadKey;
      Case Ch Of
        #59: Begin                         { F1 }
               DisplayHelp;
               Display;
             End;
        #71: TopAddr := $5000;             { Home }
        #72: Begin                         { Up }
               If TopAddr > 0 Then Dec(TopAddr);
             End;
        #73: Begin                         { PgUp }
               SDBD := False;
               I := TopAddr;
               Repeat
                 J := 0;
                 K := I;
                 While K < TopAddr Do
                 Begin
                   Inc(J);
                   Opcode := Buf[K Shr 12]^[K And $FFF] And $3FF;
                   Disassemble(Buf,K,SDBD);
                   SDBD := (Opcode = 1);
                 End; { While }
                 If J < _DLines - 1 Then Dec(I);
               Until J >= _Dlines - 1;
               TopAddr := I;
             End;
        #79: TopAddr := $FFFF;             { End }
        #80: Begin                         { Down }
               I := TopAddr;
               Opcode := Buf[TopAddr Shr 12]^[TopAddr And $FFF] And $3FF;
               Disassemble(Buf,TopAddr,SDBDInst[0]);
             End;
        #81: TopAddr := Addr;              { PgDn }
        #82: TopAddr := Regs[7];           { Insert }
      End; { Case }
      Ch := #0;
    End;
  Until Ch In [#27,'S','T','H','G'];
  Dispose(App,Done);
  MouseClose;
  While Keypressed Do ReadKey;
  If HiRes Then Set320By200Mode;
  RestoreGameScreen;
  Debug := (Ch In ['S','T','H','G']);
End; { Debug }

Begin
  VDelay       := 0;
  HDelay       := 0;
  VertOffset   := 4;
  BreakExpColl := Nil;
  Debugger     := False;
  BreakExp     := '';
  Win          := Nil;
  _WindowList  := Nil;
  _ListWork    := Nil;
  DataTop      := 0;
End.